# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

import abc
from dataclasses import dataclass

import torcharrow.dtypes as dt

from .icolumn import Column

# -----------------------------------------------------------------------------
# MapColumn


class MapColumn(Column):
    def __init__(self, device, dtype):
        assert dt.is_map(dtype)
        super().__init__(device, dtype)
        # must be set by subclasses
        self.maps: MapMethods = None


# -----------------------------------------------------------------------------
# MapMethods


class MapMethods(abc.ABC):
    """Vectorized list functions for ListColumn"""

    def __init__(self, parent):
        self._parent: MapColumn = parent

    @abc.abstractmethod
    def keys(self):
        """
        Return a list of keys for each map entry.

        See Also
        --------
        maps.values - Returns list of values for each map entry.

        Examples
        --------
        >>> import torcharrow as ta
        >>> mf = ta.column([
        >>>  {'helsinki': [-1.3, 21.5], 'moscow': [-4.0,24.3]},
        >>>  {'algiers':[11.2, 25.2], 'kinshasa':[22.2,26.8]}
        >>>  ])
        >>> mf.maps.keys()
        0  ['helsinki', 'moscow']
        1  ['algiers', 'kinshasa']
        dtype: List(string), length: 2, null_count: 0
        """
        pass

    @abc.abstractmethod
    def values(self):
        """
        Return a list of values for each map entry.

        See Also
        --------
        maps.keys - Returns list of keys for each map entry.

        Examples
        --------
        >>> import torcharrow as ta
        >>> mf = ta.column([
        >>>  {'helsinki': [-1.3, 21.5], 'moscow': [-4.0,24.3]},
        >>>  {'algiers':[11.2, 25.2], 'kinshasa':[22.2,26.8]}
        >>>  ])
        >>> mf.maps.values()
        0  [[-1.3, 21.5], [-4.0, 24.3]]
        1  [[11.2, 25.2], [22.2, 26.8]]
        dtype: List(List(float64)), length: 2, null_count: 0
        """
        pass

    def get(self, i, fill_value):
        self._parent._prototype_support_warning("maps.get")

        me = self._parent

        def fun(xs):
            # TODO improve perf by looking at lists instead of first building a map
            return xs.get(i, fill_value)

        return me._vectorize(fun, me.dtype.item_dtype)


# ops on maps --------------------------------------------------------------
#  'get',
#  'items',
#  'keys',
#  'pop',
#  'popitem',
#  'setdefault',
#  'update',
#  'values'
