# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from typing import Optional, Sequence, Union

import torcharrow.dtypes as dt

from .icolumn import Column
from .idataframe import DataFrame
from .interop_arrow import _from_arrow_array, _from_arrow_table
from .scope import Scope


def from_arrow(
    data, dtype: Optional[dt.DType] = None, device: str = ""
) -> Union[Column, DataFrame]:
    """
    Convert arrow array/table to a TorchArrow Column/DataFrame.
    """
    import pyarrow as pa

    assert isinstance(data, pa.Array) or isinstance(data, pa.Table)

    device = device or Scope.default.device

    if isinstance(data, pa.Array):
        return _from_arrow_array(data, dtype, device=device)
    elif isinstance(data, pa.Table):
        return _from_arrow_table(data, dtype, device=device)
    else:
        raise ValueError


def from_pandas(data, dtype: Optional[dt.DType] = None, device: str = ""):
    """
    Convert Pandas series/dataframe to a TorchArrow Column/DataFrame.
    """
    raise NotImplementedError


def from_pysequence(
    data: Sequence, dtype: Optional[dt.DType] = None, device: str = ""
) -> Column:
    """
    Convert Python sequence of scalars or containers to a TorchArrow Column/DataFrame.
    If dtype is None, it will be automatically inferred from data when possible.
    """

    if not dtype:
        dtype = dt.infer_dtype_from_prefix(data)

    device = device or Scope.default.device

    # pyre-fixme[6]: For 2nd param expected `DType` but got `Optional[DType]`.
    return Scope.default._FromPySequence(data, dtype, device)
